// Copyright (c) 2015, Electric Cloud, Inc.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   - Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
//   - Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// (As an historical note, the above license instantiates
// the "BSD 2-Clause License" template that was presented
// by the Open Source Initiative at
//
//   http://www.opensource.org/licenses/bsd-license.php
//
// on 2012-02-23.)

// Summary:
//
// This header defines general utilities for
// wrapping vrpsx interfaces in C++ constructs.
//
// NOTE: This convenience header does not affect the
// definition of any any binary interface; see README.txt.
//
// OPEN QUESTION:
//
// Currently we define a hierarchy of C++ wrapper classes that cannot
// be directly instantiated, and use pointer casting to convert to and
// from the C interface types.
//
// This design allows us to use C++ inheritance to express the case in
// which one interface extends another.  (It also lets us to convert
// pointers to pointers; for example, see the partial specialization
// of c_type_for_t.)
//
// But this design does not permit us to make a C++ wrapper be a base
// class of an implementation of its interface, though we can still
// use the C interface type as a base struct.
//
// Would it be better to give up the direct representation of the
// interface hierarchy in order to allow derivation from C++ wrappers
// as a way to implement C interfaces?  As yet we have said "no" because
// clients are more common than implementors, but we could reconsider.

#ifndef VRPSX_CPP_WRAPPER_HPP
#define VRPSX_CPP_WRAPPER_HPP 1

#include "../utility/vrpsx_preprocessor.h"
#include <boost/utility.hpp>


namespace vrpsx {

    // c_type_for_t< Wrapper >::type is a typedef for the
    // C interface type corresponding to C++ wrapper type Wrapper.
    //
    // To specialize, consider using VRPSX_CPP_WRAPPER.
    //
    // See also "OPEN QUESTION", above.

    template< typename Wrapper >
    struct c_type_for_t;

    // cpp_type_for_t< Interface >::type is a typedef
    // for the C++ wrapper type corresponding
    // to the C interface type Interface.
    //
    // To specialize, consider using VRPSX_CPP_WRAPPER.
    //
    // See also "OPEN QUESTION", above.

    template< typename Interface >
    struct cpp_type_for_t;

    // Because we are not changing the actual pointer value when we
    // convert between C and C++ types, we can also convert pointers
    // to such pointers.

    template< typename Wrapper >
    struct c_type_for_t< Wrapper * >
    { typedef typename c_type_for_t< Wrapper >::type *type; };

    template< typename Interface >
    struct cpp_type_for_t< Interface * >
    { typedef typename cpp_type_for_t< Interface >::type *type; };

    // Declares the given identifier to be a C++ class that acts as
    // a wrapper for the C interface having the corresponding name.
    //
    // Note that you can convert between C interface pointers
    // and C++ wrapper pointers without definitions of either
    // type--mere declarations suffice.

#define VRPSX_CPP_WRAPPER(WRAPPER)                                      \
    class WRAPPER;                                                      \
                                                                        \
    template< > struct c_type_for_t< WRAPPER >                          \
    { typedef VRPSX_CAT2(vrpsx_,WRAPPER) type; };                       \
                                                                        \
    class WRAPPER;                                                      \
    template< > struct cpp_type_for_t< VRPSX_CAT2(vrpsx_,WRAPPER) >     \
    { typedef WRAPPER type; }


    // Converts from a C interface to a C++ wrapper.

    template< typename Interface >
    static typename cpp_type_for_t< Interface >::type *to_cpp(
        Interface *iface)
    {
        return reinterpret_cast<
            typename cpp_type_for_t< Interface >::type * >(iface);
    }

    template< typename Interface >
    static typename cpp_type_for_t< Interface >::type const *to_cpp(
        Interface const *iface)
    {
        return reinterpret_cast<
            typename cpp_type_for_t< Interface >::type const * >(iface);
    }


    // Converts from a C++ wrapper to a C interface.

    template< typename Wrapper >
    static typename c_type_for_t< Wrapper >::type *from_cpp(
        Wrapper *wrapper)
    {
        return reinterpret_cast<
            typename c_type_for_t< Wrapper >::type * >(wrapper);
    }

    template< typename Wrapper >
    static typename c_type_for_t< Wrapper >::type const *from_cpp(
        Wrapper const *wrapper)
    {
        return reinterpret_cast<
            typename c_type_for_t< Wrapper >::type const * >(wrapper);
    }


// For the frequent case in which the vtbl of an interface
// includes a constant of a small type we can return by value.

#define VRPSX_CONSTANT_METHOD(TYPE, NAME)                               \
    TYPE NAME() const { return from_cpp(this) -> vtbl -> NAME; }

}


#endif /* VRPSX_CPP_WRAPPER_HPP */
