/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares an implementation of a virtual POSIX-like interface on top
 * of a POSIX-like system.  Also declares the trivial embedding thereof.
 */

#ifndef VRPSX_DIRECT_IMPLEMENTATION_H
#define VRPSX_DIRECT_IMPLEMENTATION_H 1

#include <vrpsx_argv.h>
#include <vrpsx_embedding.h>
#include <stdio.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* Given the addresses of two interface identifiers,
 * compares those identifiers by value.
 */
extern int direct_interface_identifier_compare(
    void const *va,
    void const *vb);

extern vrpsx_errno_t direct_embedding_acquire(
    vrpsx_embedding_t **embedding);

extern void direct_error_create(
    vrpsx_errno_t ern,
    vrpsx_error_t **out_error);

/* Call this method *before* the first call to direct_embedding_acquire.
 * Calling it afterward has various unsafe implications, despite locking.
 */
extern void direct_trace(
    FILE *new_trace_file);

/* While emitting a sequence of values in a trace message,
 * this struct indicates how to handle the next value.
 */
typedef struct trace_state_struct {
    /* The prefix before emitting the next value.
     */
    char const *prefix;

    /* Text to emit to close the current sequence of values.
     */
    char const *closer;

    /* After emitting a value, transition to this state.
     */
    struct trace_state_struct const *next;
} trace_state_t;

/* While emitting a sequence of values in a trace message, stacks with
 * this node type track the syntactic state as a stack of trace_state_t.
 *
 * The trace lock is held by the thread owning a nonempty stack.
 */
typedef struct trace_stack_struct {
    /* The directly containing syntactic state.
     */
    struct trace_stack_struct *outer;

    /* The current state at this level of the stack.
     */
    trace_state_t const *current;
} trace_stack_t;

/* Constructs "*inner" with the given member values, but also:
 *
 * If outer != NULL, then this function emits outer->current->prefix.
 *
 * Otherwise, it locks the trace mutex and calls trace_thread_prefix().
 *
 * After using "*inner", please call "trace_end(inner)".
 */
extern void trace_begin(
    trace_stack_t *inner,
    trace_stack_t *outer,
    trace_state_t const *first);

/* Emits inner->current->closer and then destroys "*inner".
 *
 * Then transitions the "current" member of the immediately containing
 * trace_stack_t, but if there is no such instance, then flushes the
 * trace file and then unlocks the trace mutex.
 */
extern void trace_end(
    trace_stack_t *inner);

/* Starts to trace entry into an interface call with the specified name.
 *
 * LOCKS the trace mutex, but does NOT unlock it.
 *
 * After emitting the arguments, please call "trace_end(inner)".
 */
extern void trace_enter_begin(
    trace_stack_t *inner,
    char const *func);

/* Starts to trace departure from an interface call.
 *
 * LOCKS the trace mutex, but does NOT unlock it.
 *
 * After emitting the returned value(s), please call "trace_end(inner)".
 */
extern void trace_leave_begin(
    trace_stack_t *inner,
    char const *func);

/* Equivalent to:
 *
 *    trace_leave_begin(inner, func);
 *    trace_system_error(inner, ern);
 *
 * This function is merely a convenience.
 */
extern void trace_leave_begin_ern(
    trace_stack_t *inner,
    char const *func,
    vrpsx_errno_t ern);

/* Finishes tracing of departure from an interface call
 * that has one or more output arguments.
 *
 * If (trace_separator != trace_outputs_opener),
 * then this function emits a close parenthesis.
 *
 * This function sets trace_separator an the empty separator.
 *
 * UNLOCKS the trace mutex, WITHOUT first locking it.
 */
extern void trace_leave_end(void);

/* Starts to trace a general-purpose note that
 * does not signify an interface entry or return.
 *
 * LOCKS the trace mutex, but does NOT unlock it.
 *
 * After emitting values of interest, please call "trace_end(inner)".
 */
extern void trace_note_begin(
    trace_stack_t *inner,
    char const *name);

/* Traces a system error message number, and if that
 * number is nonzero, follows it with the error text.
 *
 * You can pass an error number from the underlying POSIX-like system,
 * or an error number from vrpsx_system_t / vrpsx_embedding_t, because
 * in this implementation there is no conflict between those two spaces.
 */
extern void trace_system_error(
    trace_stack_t *stack,
    vrpsx_errno_t ern);

/* Traces an int in decimal.
 */
extern void trace_decimal_int(
    trace_stack_t *stack,
    int value);

/* Traces a representation of the NULL pointer (currently "NULL").
 */
extern void trace_null(
    trace_stack_t *stack);

/* Prints a C string literal representation of the given argument string,
 * unless (str == NULL), in which case the output is "NULL" (without quotes).
 */
extern void print_string(
    FILE *fp,
    char const *str);

/* Traces a C string literal representation of the given argument string,
 * unless the argument is NULL, in which case the effect is the same as
 * if you had called "trace_null(stack)".
 */
extern void trace_string(
    trace_stack_t *stack,
    char const *str);

/* Traces an argument expressing the given memory address.
 */
extern void trace_address(
    trace_stack_t *stack,
    void const *ptr);

/* Traces a uintptr_t in hexadecimal.
 */
extern void trace_uintptr(
    trace_stack_t *stack,
    uintptr_t value);

/* Traces a C identifier.
 */
extern void trace_identifier(
    trace_stack_t *stack,
    char const *id);

/* Traces an interface identifier.
 */
extern void trace_interface_identifier(
    trace_stack_t *stack,
    vrpsx_interface_identifier_t id);

/* Defines a named flag, typically as part
 * of an array terminated by { 0, 0, NULL }.
 *
 * Used to decode a flag set created by the
 * bitwise OR of various named bit masks.
 */
typedef struct flag_definition_struct {
    vrpsx_uint64_t mask;    // If this subset of bits...
    vrpsx_uint64_t value;   // ...has this value, then...
    char const *name;       // ...show this name when describing the flag set.
} flag_definition_t;

/* Traces a set of flags as a bitwise OR.
 *
 * The operands of that '|' include the names of the defined flags
 * whose masks are fully included in the given set, together with
 * a hexadecimal number if some bits remain anonymous.
 */
extern void trace_flagset(
    trace_stack_t *stack,
    flag_definition_t const *flagdefs,
    vrpsx_uint64_t flags);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_DIRECT_IMPLEMENTATION_H */
