/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Implements doubly-linked lists.
 */

#define __STRICT_ANSI__ 1

#include "test_dlist.h"
#include <stddef.h>


extern void dlink_construct(
    dlink_t *link)
{
    link->prev = link;
    link->next = link;
}

extern void dlink_destroy(
    dlink_t *link)
{
    dlink_t *prev = link->prev;
    dlink_t *next = link->next;

    prev->next = next;
    next->prev = prev;
}

extern void dlink_splice(
    dlink_t *beforeA,
    dlink_t *beforeB)
{
    dlink_t *afterA = beforeA->next;
    dlink_t *afterB = beforeB->next;

    beforeA->next = afterB;
    afterB->prev = beforeA;

    beforeB->next = afterA;
    afterA->prev = beforeB;
}

static dlink_t *dlink_as_slink_sort(
    dlink_t *slist,
    int (*compare)(void *, dlink_t const *, dlink_t const *),
    void *callback_data)
{
    dlink_t *head0, *head1, *resultHead;
    dlink_t **tail0, **tail1, **resultTail;

    /* Lists of length 0 or 1 are already sorted.
     */
    if (! slist || ! slist->next) {
        return slist;
    }

    /* The list contains at least 2 nodes.  Split the list into two
     * lists whose lenghts are as close to being equal as is possible.
     */
    tail0 = &head0;
    tail1 = &head1;

    while (slist) {
        *tail0 = slist;
        slist = *(tail0 = &slist->next);

        if (slist) {
            *tail1 = slist;
            slist = *(tail1 = &slist->next);
        }
    }

    *tail0 = NULL;
    *tail1 = NULL;

    /* Recursively sort the two lists.
     */
    head0 = dlink_as_slink_sort(head0, compare, callback_data);
    head1 = dlink_as_slink_sort(head1, compare, callback_data);

    /* Merge the two lists back together.
     */
    resultTail = &resultHead;

    for (;;) {
        if ((*compare)(callback_data, head0, head1) <= 0) {
            *resultTail = head0;
            head0 = *(resultTail = &head0->next);

            if (! head0) {
                *resultTail = head1;
                return resultHead;
            }
        } else {
            *resultTail = head1;
            head1 = *(resultTail = &head1->next);

            if (! head1) {
                *resultTail = head0;
                return resultHead;
            }
        }
    }
}

extern void dlink_sort(
    dlink_t *term,
    int (*compare)(void *, dlink_t const *, dlink_t const *),
    void *callback_data)
{
    dlink_t *prev, *node;

    /* For speed and simplicity, temporarily treat this
     * doubly-linked list as just a singly-linked list.
     */
    term->prev->next = NULL;

    term->next = dlink_as_slink_sort(term->next, compare, callback_data);

    /* Convert the sorted singly-linked list back into a doubly-linked list.
     */
    prev = term;
    node = term->next;
    while (node) {
        node->prev = prev;
        node = (prev = node)->next;
    }
    term->prev = prev;
    prev->next = term;
}
