/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Implements utilties for timing tests and reporting those figures.
 */

#define _POSIX_C_SOURCE 200809L

#include "test_timer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define MALLOC(TYPE)  ((TYPE *)malloc(sizeof(TYPE)))


typedef struct test_timing_struct {
    dlink_t element;
    char const *name;
    struct timespec elapsed;
} test_timing_t;


extern void test_timer_construct(
    test_timer_t *obj)
{
    dlink_construct(&obj->timings);
}

extern void test_timer_destroy(
    test_timer_t *obj)
{
    while (obj->timings.next != &obj->timings) {
        test_timing_t *doomed = (test_timing_t *)obj->timings.next;
        dlink_destroy(&doomed->element);
        free(doomed);
    }

    dlink_destroy(&obj->timings);
}

extern void test_timer_add(
    test_timer_t *obj,
    char const *name,
    struct timespec const *elapsed)
{
    test_timing_t *timing = MALLOC(test_timing_t);

    if (! timing) {
        perror("malloc");
        exit(1);
    }

    dlink_construct(&timing->element);
    timing->name = name;
    timing->elapsed = *elapsed;

    dlink_splice(&obj->timings, &timing->element);
}

extern void test_timer_add_difference(
    test_timer_t *obj,
    char const *name,
    struct timespec const *begin,
    struct timespec const *end)
{
    struct timespec elapsed;

    elapsed.tv_sec = end->tv_sec - begin->tv_sec;
    elapsed.tv_nsec = end->tv_nsec - begin->tv_nsec;

    if (elapsed.tv_nsec < 0) {
        elapsed.tv_nsec += 1000000000;
        elapsed.tv_sec -= 1;
    }

    test_timer_add(obj, name, &elapsed);
}

/* ASSUMES that the given dlink_t instances are actually the "element"
 * members of test_timing_t instances, then orders those instances
 * first by decreasing elapsed time and then by name.
 */
static int test_timing_compare(
    void *unused,
    dlink_t const *nodeA,
    dlink_t const *nodeB)
{
    test_timing_t const *timingA = (test_timing_t const *)nodeA;
    test_timing_t const *timingB = (test_timing_t const *)nodeB;

    if (timingA->elapsed.tv_sec > timingB->elapsed.tv_sec) {
        return -1;
    }

    if (timingA->elapsed.tv_sec < timingB->elapsed.tv_sec) {
        return +1;
    }

    if (timingA->elapsed.tv_nsec > timingB->elapsed.tv_nsec) {
        return -1;
    }

    if (timingA->elapsed.tv_nsec < timingB->elapsed.tv_nsec) {
        return +1;
    }

    return strcmp(timingA->name, timingB->name);
}

extern void test_timer_report(
    test_timer_t *obj)
{
    dlink_t *node;

    dlink_sort(&obj->timings, &test_timing_compare, NULL);

    puts("elapsed:");

    for (node = obj->timings.next; node != &obj->timings; node = node->next) {
        test_timing_t const *timing = (test_timing_t const *)node;

        printf("\t%ld.%09ld\t%s\n",
                (long)timing->elapsed.tv_sec,
                (long)timing->elapsed.tv_nsec,
                timing->name);
    }
}
