/* Copyright (c) 2015, Electric Cloud, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* (As an historical note, the above license instantiates
 * the "BSD 2-Clause License" template that was presented
 * by the Open Source Initiative at
 *
 *   http://www.opensource.org/licenses/bsd-license.php
 *
 * on 2012-02-23.)
 */

/* Summary:
 *
 * Declares an interface through which you can allocate and deallocate memory.
 */

#ifndef VRPSX_ALLOCATOR_H
#define VRPSX_ALLOCATOR_H 1

#include <stddef.h> /* EXCLUDE_FROM_INTERFACE_IDENTIFIER because basic
                     * definitions such as size_t are part of the architecture,
                     * and we are not attempting binary compatibility between
                     * different architectures (such as calling from 32-bit
                     * code into 64-bit code).
                     */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* An interface through which you can allocate and deallocate memory.
 *
 * Interface operations that accept parameters of this type implicitly extend
 * this interface by requiring the allocated blocks to be aligned in whatever
 * manner is required by the data to be stored in the allocated block.  If it
 * is not obvious what that requirement is, then the operation must document
 * its alignment requirement.
 *
 * In the absence of more specific documentation, the default usage pattern
 * is for a function to take an interface of this type and use it only until
 * it returns.  More complex cases are certainly possible, especially with
 * those extensions of this interface that add lifetime controls.
 *
 * Singular Allocation:
 *
 * When a function takes an interface of this type as an argument, it may make
 * the "singular allocation" guarantee for that argument: namely, that it will
 * use that argument to allocate at most one block at a time, and that any
 * block still allocated upon return will have exactly the required size.
 *
 * The purpose of this guarantee is to allow the caller to optimize allocation
 * in various ways.  For example, the caller may be able to supply the ultimate
 * location where the result will be kept, instead of having to copy it after
 * the present function returns.  Or perhaps the caller needs the result only
 * temporarily, and wants to use a stack buffer for small results without the
 * complexity and run-time cost of handling multiple simultaneous allocations.
 *
 * Now, a function making the "singular allocation" guarantee for one argument
 * may nevertheless allocate additional memory blocks through other allocator
 * arguments (even if they happen to have the same pointer value in the present
 * call) or from other available resources (such as the system allocator).
 *
 * Furthermore, a function making the "singular allocation" guarantee for an
 * argument may free a block that it allocated using that argument and then
 * use the same allocator to allocate another block.  But please minimize such
 * reallocations because the allocator may place the requested block inside of
 * some larger structure, in order to avoid later copying as mentioned above.
 * We allow such reallocation only because it is unavoidable where the result
 * size changes after allocation begins, due to concurrency (the actions of
 * another thread or process) or reentrancy (the actions of the allocator).
 *
 * NUL-Terminated Allocation:
 *
 * When a function takes an interface of this type as an argument, it may make
 * the "NUL-terminated allocation" guarantee for that argument: namely, that if
 * it modifies the final char in an allocation, then before any free, return,
 * or execution of a client-supplied callback, it will ensure that that final
 * char has value '\0'.  But it may temporarily write some other value,
 * and is not obligated to write a value at all to that particular char.
 *
 * C++ Implementations:
 *
 * Guaranteeing both "singular allocation" and "NUL-terminated allocation"
 * in the context of allocating a NUL-terminated string enables efficient
 * implementation of this interface on top of a C++ std::string whose "c_str()"
 * method will then return the allocated string.  In particular, the size
 * of the std::string may be the "strlen" of the allocated C-style string,
 * with the NUL terminator at the following position, so that there is no
 * need to copy or resize a result before it can be used as a std::string.
 *
 * Interface Lifetime: implied by the context, or by an interface extension.
 * Thread Safety: implied by the context.
 */
typedef struct vrpsx_allocator_struct {
#define VRPSX_ALLOCATOR_MEMBERS(VTBL)          \
    VTBL const *vtbl;

    VRPSX_ALLOCATOR_MEMBERS(struct vrpsx_allocator_vtbl_struct)
} vrpsx_allocator_t;

typedef struct vrpsx_allocator_end_struct {
} vrpsx_allocator_end_t;

typedef struct vrpsx_allocator_vtbl_struct {
#define VRPSX_ALLOCATOR_VTBL_MEMBERS(TYPE)                              \
    /* Deallocates memory allocated by the "allocate_memory" operation  \
     * of this interface.                                               \
     *                                                                  \
     * Does nothing if you pass NULL.                                   \
     */                                                                 \
    void (*free_memory)(                                                \
        TYPE *allocator,                                                \
        void *memory);                                                  \
                                                                        \
    /* Allocates a block of memory of the given size and returns its address. \
     * Returns NULL on failure.                                         \
     *                                                                  \
     * You should eventually call "free_memory" to deallocate any memory \
     * allocated by this operation.  Pass the same allocator interface. \
     *                                                                  \
     * The alignment of allocated memory is implementation-specified,   \
     * but see the introductory comments for information on how this    \
     * interface is implicitly affected by alignment requirements.      \
     *                                                                  \
     * If (size == 0), then the implementation may choose to return NULL. \
     */                                                                 \
    void *(*allocate_memory)(                                           \
        TYPE *allocator,                                                \
        size_t size);                                                   \
                                                                        \
    /* End marker for documentation and to help detect initialization bugs. \
     */                                                                 \
    vrpsx_allocator_end_t const *allocator_end;

    VRPSX_ALLOCATOR_VTBL_MEMBERS(vrpsx_allocator_t)
} vrpsx_allocator_vtbl_t;


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* VRPSX_ALLOCATOR_H */
